@@ -7,6 +7,7 @@
 # License:: Ruby license.
 
 require 'test/unit/color-scheme'
+require 'test/unit/code-snippet-fetcher'
 require 'test/unit/ui/testrunner'
 require 'test/unit/ui/testrunnermediator'
 require 'test/unit/ui/console/outputlevel'
@@ -44,6 +45,7 @@ module Test
             @top_level = true
             @current_output_level = NORMAL
             @faults = []
+            @code_snippet_fetcher = CodeSnippetFetcher.new
           end
 
           private
@@ -170,39 +172,51 @@ module Test
                 fault.inspected_expected and fault.inspected_actual
               output_single(fault.label, fault_color(fault))
               output(":")
+              output(fault.test_name)
               output_fault_backtrace(fault)
               output_fault_message(fault)
             else
-              label, detail = format_fault(fault).split(/\r?\n/, 2)
-              output(label, fault_color(fault))
-              output(detail)
+              output_single(fault.label, fault_color(fault))
+              if fault.is_a?(Error)
+                output(": #{fault.test_name}")
+                output(fault.message)
+              else
+                output(": #{fault.message}")
+                output(fault.test_name)
+              end
+              output_fault_backtrace(fault)
             end
           end
 
           def output_fault_backtrace(fault)
             backtrace = fault.location
-            if backtrace.size == 1
-              output(fault.test_name +
-                     backtrace[0].sub(/\A(.+:\d+).*/, ' [\\1]') +
-                     ":")
-            else
-              output(fault.test_name)
-              backtrace.each_with_index do |entry, i|
-                if i.zero?
-                  prefix = "["
-                  postfix = ""
-                elsif i == backtrace.size - 1
-                  prefix = " "
-                  postfix = "]:"
-                else
-                  prefix = " "
-                  postfix = ""
-                end
-                output("    #{prefix}#{entry}#{postfix}")
-              end
+            backtrace.each_with_index do |entry, i|
+              output(entry)
+              output_code_snippet(entry, fault_color(fault)) if i.zero?
             end
           end
 
+          def output_code_snippet(entry, target_line_color=nil)
+            return unless /\A(.*):(\d+)/ =~ entry
+            file = $1
+            line_number = $2.to_i
+            lines = @code_snippet_fetcher.fetch(file, line_number)
+            max_n = lines.collect {|n, line, attributes| n}.max
+            digits = (Math.log10(max_n) + 1).truncate
+            lines.each do |n, line, attributes|
+              if attributes[:target_line?]
+                line_color = target_line_color
+                current_line_mark = "=>"
+              else
+                line_color = nil
+                current_line_mark = ""
+              end
+              output("  %2s %*d: %s" % [current_line_mark, digits, n, line],
+                     line_color)
+            end
+            nl
+          end
+
           def output_fault_message(fault)
             if fault.expected.respond_to?(:encoding) and
                 fault.actual.respond_to?(:encoding) and
